"""Various tools for electomagntism physics
Variables comonly used in the package
-------------------------------------

bmw = Beamwidth angle
bw = Bandwidth
cl : Correlation length
epsilon, ep : Electric permittivity
h = Altitude
mp : Magnetic permeability
nRCS : Normalized Radar Cross Section
RCS : Radar Cross Section
n : Optical index
R : Reflection coefficient
sigma_h, sh : Root Mean Square height
theta, th : angle
T : Transmission coefficient
wf : Wave frequency
wk : Wave number
wl : Wavelength
"""

__version__ = "1.1.4"
__author__ = "Cyril Grima"

__all__ = ["iem", "invert", "roughness", "surface", "utils", "filtering", "simulation", "fractal"]

from . import iem, invert, roughness, surface, utils, filtering, simulation, fractal



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from scipy import signal 
from scipy import fft
import scipy.constants as ct
from scipy import integrate
from scipy import special
from scipy import constants
import subradar as sr
import pandas as pd
import glob
from labellines import labelLine, labelLines
import seaborn as sns

# -------------------------------
# TODO: INTEGRATION WITH THE REPO
# -------------------------------


def dB(func):
    """Decorator to convert a linear power to dB power
    """
    def func_wrapper(*args, dB=True, **kwargs):
        out = func(*args, **kwargs)
        if dB is True:
            return 10*np.log10(out)
        else:
            return out
    return func_wrapper


def deg2frac(x):
    """Slope Degree to fraction conversion
    """
    return np.tan(np.deg2rad(x))


def frac2deg(x):
    """Fraction to Slope degree conversion
    """
    return np.rad2deg(np.arctan(x))


class Scattering:
    """Fractal Model"""
    def __init__(self, 
                 wf = None, # Wave frequency
                 wb = None, # Bandwidth
                 Z = None, # Surface altitude
                 Re = None, # Effective reflectivity
                 H = None, # Husrt exponent
                 RMSh_1 = None, # RMS Height at 1m
                 phi = 0. # Backscattering angle
                ):
        self.wf = wf
        self.wb = wb
        self.Z = Z
        self.Re = Re
        self.H = H
        self.RMSh_1 = RMSh_1
        self.phi = phi
        self.wl = constants.c/self.wf

    def RMSh(self, scale=None):
        """RMS Height at the chosen scale (scale = wavelength by default)
        [Shepard et al, 1999(2a)]
        """
        if scale is None:
            scale = self.wl
        out = self.RMSh_1*(scale)**self.H
        return out
    
    def RMSd(self, scale=None):
        """RMS Derivation at the chosen scale (scale = wavelength by default)
        """
        if scale is None:
            scale = self.wl
        out = self.RMSh(scale=scale)*np.sqrt(2)
        return out
    
    def RMSs(self, scale=None, deg=False):
        """RMS slope at the chosen scale (scale = wavelength by defautl)
        [Shepard et al, 1999(1c)]
        """
        if scale is None:
            scale = self.wl
        out = self.RMSd(scale=scale)/scale
        if deg is True:
            return np.rad2deg(np.arctan(out))
        return out
    
    @dB
    def geometric_loss(self):
        """1-way loss due to geometric spreading"""
        out = (1/4/np.pi/self.Z**2)
        return out
    
    @dB
    def pulselimited_footprint(self, area=True, norm=True):
        """Pulse-limited footprint radius OR area
        """
        out = np.sqrt(constants.c*self.Z/self.wb) # Radius
        if area is True:
            out = np.pi*out**2 # Area
        if norm is True:
            out = out/self.wl
        return out
    
    @dB
    def fresnel_zone(self, area=True, norm=True):
        """Fresnel zone radius OR area
        """
        out = np.sqrt(self.Z*self.wl/2)
        if area is True:
            out = np.pi*out**2 # Area
        if norm is True:
            out = out/self.wl
        return out
    
    def coherent_effective_radius(self, n=5, norm=True):
        """Effective radius relative to wavelength
        [Shepard et al, 1999(20)]
        """
        frac = n/(4*np.pi**2*self.RMSs()**2*np.cos(self.phi)**2)
        out = (frac)**(1/2/self.H)
        if norm is not True:
            out = out*self.wl
        return out
    
    def incoherent_limit_radius(self, norm=True):
        """Radius below which the incoherent model is not valid
        """
        pass
    
    @dB
    def flat_RCS(self, norm=False, Re=1.):
        """Perfect-reflector normalized radar cross-section
        """
        #A = self.fresnel_zone(area=True, dB=False)
        #out = 4*np.pi*A**2/self.wl
        out = self.coherent_RCS(Re=Re, H=0., sl=0., norm=False, dB=False)
        if norm is True:
            out = out/A
        return out
    
    
    @dB
    def coherent_RCS(self, model='Campbell2003', norm=False,
                    Re=None, H=None, sl=None, phi=None, rmax=None):
        """Coherent normalized radar cross-section
        """
        # Parameters
        if Re is None:
            Re = self.Re 
        if sl is None:
            sl = self.RMSs()
        if H is None:
            H = self.H
        if phi is None:
            phi = self.phi
        if rmax is None:
            #rmax = self.pulselimited_footprint(area=False, dB=False)
            rmax = self.fresnel_zone(area=False, dB=False, norm=True)
            #rmax = self.coherent_effective_radius()
        
        if model == 'Campbell2003':
            _integrand = lambda x: x * np.exp(-4*np.pi**2*sl**2*x**(2*H)*\
                         np.cos(phi)**2)*special.j0(4*np.pi*x*np.sin(phi))
            K, err = integrate.quad(_integrand, 0, rmax)
            #out = 16*(np.pi*Re*K/rmax)**2 #nRCS
            out = 16*np.pi**3*self.wl**2*Re**2*K**2 #RCS
            
        if norm is True:
            out = out/(np.pi*rmax**2*self.wl**2)
            
        return out

    @dB
    def incoherent_RCS(self, model='Biccari2001', norm=False, 
                      Re=None, H=None, sl=None, phi=None, rmax=None):
        """Incoherent normalized radar cross-section
        """
        # Parameters
        if Re is None:
            Re = self.Re 
        if sl is None:
            sl = self.RMSs()
        if H is None:
            H = self.H
        if phi is None:
            phi = self.phi
        if rmax is None:
            rmax = self.pulselimited_footprint(area=False, dB=False)
        
        if model == 'Biccari2001': #nRCS
            if (H == 0) or (sl == 0):
                return 0.
            elif phi == 0.:
                K = special.gamma(1/H)
            else:
                K = nsum(lambda n: 
                         (-1)**n*np.sin(phi)**(2*n)/\
                         np.math.factorial(n)**2*\
                         ((2*np.pi)**(H-1))**(2*n/H)/\
                         (sl*np.sqrt(2)*np.cos(phi))**(2*n/H)*\
                         np.math.factorial((n+1)/H-1),\
                         [0, inf])
            try:
                out = K*Re**2/H* ((2*np.pi)**(H-1)/sl/np.sqrt(2))**(2/H)
            except ZeroDivisionError:
                out = 0
            out = np.float64(out)
            
        out_RCS = out*(np.pi*rmax)**2*self.wl**2
            
        if norm is False:
            out = out_RCS
        
        if out_RCS > self.flat_RCS(Re=self.Re, dB=False, norm=False):
            out = 0
            
        return out
    
    @dB
    def total_RCS(self):
        """Total normalized radar cross-section
        """
        _sc = self.coherent_RCS(norm=False, dB=False)
        _sn = self.incoherent_RCS(norm=False, dB=False)
        if _sn > self.flat_RCS(norm=False, dB=False, Re=self.Re):
            out = _sc
        else:
            out = _sc + _sn
        return out


"""Tools to use the Integral Equation Method (IEM) in the monostatic
(i.e. backscattering) case, when u = k*sin(th), case as described in
Fung et al. [1994].

Note: Argument "m" is subradar.Model class
"""

import numpy as np
from numpy import cos, exp, float64, inf, log10, sin
from scipy.special import factorial
from mpmath import nsum

from . import roughness
from .Classdef import Fresnel, Roughness, Signal


nan = float('nan')


class Common(Fresnel):
    """Common IEM relationships"""
    def __init__(self, **kwargs):
        Fresnel.__init__(self, **kwargs)

    def f(self):
        """f_pp terms"""

        return {'vv': 2*self.R['vv']/cos(self.th),
                'hh':-2*self.R['hh']/cos(self.th)}

    def ft(self):
        """f_ppt terms
        !! FOR NORMAL INCIDENCE ONLY !!
        """

        nr = np.sqrt(self.mu/self.ep)

        vv = -(1-self.R['vv']) - (1+self.R['vv'])*nr #eq 4D.1
        hh = +(1+self.R['hh']) + (1-self.R['hh'])*nr #eq 4D.2

        return {'vv': vv, 'hh': hh}


class Large_S(Signal, Fresnel):
    """Large roughness single-scatter approximation [sec. 5.5]"""



class Large_M(Signal, Fresnel):
    """Large roughness multiple-scatter approximation [sec. 5.5]"""



class Small_S(Common, Signal, Fresnel, Roughness):
    """Small-to-moderate roughness single-scatter approximation [sec. 5.5]"""
    def __init__(self, **kwargs):
        Roughness.__init__(self, **kwargs)
        Fresnel.__init__(self, **kwargs)
        Signal.__init__(self, **kwargs)
        Common.__init__(self, **kwargs)


    def F_sum(self):
        """F_pp(-)+F_pp(+) terms"""

        vv = 2*sin(self.th)**2 * (1+self.R['vv'])**2 / cos(self.th) * \
              ( (1-1/self.ep) + \
                (self.mu*self.ep - sin(self.th)**2 - self.ep*cos(self.th)**2) /
                (self.ep*cos(self.th))**2
              )

        hh = -2*sin(self.th)**2 * (1+self.R['hh'])**2 / cos(self.th) * \
              ( (1-1/self.mu) + \
                (self.mu*self.ep - sin(self.th)**2 - self.mu*cos(self.th)**2) /
                (self.mu*cos(self.th))**2
              )
        return {'vv':vv, 'hh':hh}


    def Ft(self):
        """Ft terms
        !! FOR NORMAL INCIDENCE ONLY !!
        """

        k = self.wk
        kt = self.n2*self.wk
        R = self.R['nn']
        er = self.ep
        mr = self.mu
        nr = np.sqrt(mr/er)

        s = 0
        st = 0
        cs = 1
        cst = 1
        sf = 0
        csf = -1

        sq = np.sqrt(self.mu*self.ep)
        rem = sq
        nr = np.sqrt(self.mu/self.ep)
        a1 = (st*s-rem*csf)/(sq*cst)
        a2 = s*(st-s*csf/rem)/cst
        Tv = 1+self.R['vv']
        Tvm = 1-self.R['vv']
        Th = 1+self.R['hh']
        Thm = 1-self.R['hh']
        Tp = 1+self.R['nn']
        Tm = 1-self.R['nn']
        R = (self.R['vv']-self.R['hh'])/2

        #vv = - (cs*Tvm-sq*Tv/self.ep) * (Tv*csf+Tvm*a1) \
        #     - (Tvm**2-Tv*Tvm*cs/sq)*a2# eq 4D.41
        #hh = (cs*Thm-sq*Th/self.mu) * (Th*csf+Thm*a1)*nr \
        #     + (Thm**2-Th*Thm*cs/sq)*a2*nr # eq 4D.42

        #vv = -(Tvm-Tv*nr)*(-Tv+Tvm)
        #hh =  (Thm-Th/nr)*(-Th-Thm)*nr

        # Chris derivation (2022/09/12):

        vv = -k*(
             -(1-R)*( (1+R)/np.sqrt(k**2) + (-1+R)*er/np.sqrt(kt**2) )*nr
             +(1+R)*((-1+R)/np.sqrt(k**2) + ( 1+R)*mr/np.sqrt(kt**2) )
             )

        hh = -k*(
              (1+R)*( (1-R)/np.sqrt(k**2) - ( 1+R)*er/np.sqrt(kt**2) )*nr
             -(1-R)*(-(1+R)/np.sqrt(k**2) - (-1+R)*mr/np.sqrt(kt**2) )
             )


        return {'vv':vv, 'hh':hh}


    def I(self, n):
        """I_pp terms"""

        def _I(self, pp, n):
            return (2*self.wk_z)**n * self.f()[pp] *\
                   exp(-(self.sh*self.wk_z)**2) \
                   + self.wk_z**n * self.F_sum()[pp] /2

        return {'vv':_I(self, 'vv', n),
                'hh':_I(self, 'hh', n)}


    def It(self, n):
        """I_ppt terms
        !! FOR NORMAL INCIDENCE !!
        """

        def _It(self, pp, n):

            kz = self.wk_z
            ktz = self.n2*self.wk_z

            #return (1+self.n2)**n * self.wk_z**n * self.ft()[pp] *\
            #       exp(-self.n2*self.sh**2*self.wk_z**2) \
            #       + (1+self.n2**n) * self.wk_z**n * self.Ft()[pp]

            return (kz+ktz)**n * self.ft()[pp] *\
                   exp(-self.sh**2*kz*ktz) \
                   + (ktz**n*self.Ft()[pp] + kz**n*self.Ft()[pp] )/2

        return {'vv':_It(self, 'vv', n),
                'hh':_It(self, 'hh', n)
                }


    def nRCS(self, n='richardson+shanks', kind='isotropic gaussian', db=False,
            transmission=False):
        """Normalized Radar cross-section"""

        def _nRCS(self, pp, n, kind, transmission):
            n = float(n)

            if not transmission:
                out = self.wk**2 * exp(-2*(self.sh*self.wk_z)**2) /2 * \
                      ( self.sh**(2*n) * np.abs( self.I(n)[pp] )**2 * \
                      roughness.spectrum(self.wk, self.sh, self.cl, self.th, \
                      n=n, kind=kind) / factorial(n) )

            elif transmission:
                out = self.n2**2 * self.wk**2 * \
                      exp(-(1+self.n2**2)*self.sh**2*self.wk_z**2) /2 * \
                      ( self.sh**(2*n) * np.abs( self.It(n)[pp] )**2 * \
                      roughness.spectrum(self.wk, self.sh, self.cl, self.th, \
                      n=n, kind=kind) / factorial(n) )

            return out

        if not isinstance(n, str):
            vv = nsum(lambda x: _nRCS(self, 'vv', x, kind, transmission), [1,n] )
            hh = nsum(lambda x: _nRCS(self, 'hh', x, kind, transmission), [1,n] )
        else:
            vv = nsum(lambda x: _nRCS(self, 'vv', x, kind, transmission), [1,inf], method=n)
            hh = nsum(lambda x: _nRCS(self, 'hh', x, kind, transmission), [1,inf], method=n)

        vv, hh = float64(vv), float64(hh)

        if db:
            vv, hh = 10*log10(vv), 10*log10(hh)

        return {'vv':vv, 'hh':hh}



class Small_M(Signal, Fresnel):
    """Small-to-moderate roughness multiple-scatter approximation [sec. 5.5]"""



"""Signal inversion"""

from importlib import import_module
from scipy import integrate
import numpy as np
from numpy import cos, exp, log, log10, pi, sqrt
import matplotlib.pyplot as plt

from . import utils
from .Classdef import Fresnel, Signal



def srf2power_norminc(model, approx, gain=lambda th:1, th_max=np.nan,
                      db=True, kind='isotropic gaussian', **kwargs):
    """Power components over circular footprint at normal incident
    from surface properties

    PARAMETERS
    ==========
    gain: function
        function of the observation angle (radian) and return
        a linear amplitude
    th_max: float
        Maximum observation angle corresponding to the edge of the footprint
    **kwargs: to be passed to the model (do not include theta)

    EXAMPLE
    =======
    In[1]: sr.invert.srf2power_norminc('iem','Small_S', th_max=0.008,
           wf=13.78e9, ep2=1.5, sh=1.5e-3, cl=100e-3)
    Out[1]:
    {'pc': -23.172003557542929,
    'pn': -36.370086693995717,
    'ratio': 13.198083136452789}
    """
    m = import_module('subradar.' + model)
    Scattering = getattr(m, approx)

    # Coherent Signal
    a = Scattering(th=0, **kwargs)
    pc = a.R['nn']**2 * exp( -(2*a.wk*a.sh)**2 )

    # Incoherent Signal
    # Note: nRCS(th=0) approximation all over the footprint.
    # That allows to get nRCS out of the integral for faster computation
    nRCS = lambda th: Scattering(th=th, **kwargs).nRCS(kind=kind)['hh']
    integrand = lambda th: 2* gain(th)**2 * np.arctan(th)/(th**2+1) *nRCS(th)
    pn = integrate.quad(integrand, 0, th_max)[0]

    # Output
    ratio = pc/pn
    if db:
        pc, pn, ratio = 10*log10(pc), 10*log10(pn), 10*log10(ratio)
    return {'pc':pc, 'pn':pn, 'ratio':ratio}



def power2srf_norminc(model, approx, pc, pn, gain=lambda th:1, wf=np.nan,
              th_max=.1, db=True, kind='isotropic gaussian',
              ep_range=(1.4,2.5), cl_logrange=(-1, 2), n=50, verbose=False):
    """Surface properties solutions from Power components [in dB]

    EXAMPLE
    =======
    sr.invert.power2srf_norminc('iem','Small_S', pc, pn, th_max=3/1000.,wf=wf,
    verbose=True, cl_logrange=(5,), n=50)

    NOTE
    ====
    For large correlation length estimation, set cl_logrange to an array of
    1 element with what you think is the cl threshold. (see example above).
    Else, use an array of 2 elements to determine the range.
    """
    pc = 10**(pc/10.)
    s = Signal(wf=wf, wb=np.nan, th=th_max)

    ep = np.linspace(ep_range[0], ep_range[1], n)
    r = utils.R(1, ep, 1, 1, s.th)
    #cl = 10**np.linspace(cl_logrange[0], cl_logrange[1], n)

    sh = sqrt(log(r**2/pc)) / (2*s.wk*cos(s.th))
    sh[np.isnan(sh)] = 0

    # set the iterations for cl
    jn = 1 if len(cl_logrange) == 1 else n
    cl = 10**np.linspace(cl_logrange[0], cl_logrange[-1], jn)
    cl_out = np.full_like(ep, np.nan)

    # Iterations over the field of parameters
    for i, val in enumerate(ep):
        if verbose:
            print('\n')
        if sh[i] == 0: #if no solution for sh, do not compute
            continue
        for j in range(jn-1, -1, -1): # j in reversed(range(0, jn, 1)):
            tmp = srf2power_norminc(model, approx, gain=gain, th_max=th_max,
                  wf=wf, ep2=ep[i], sh=sh[i], cl=cl[j])['pn']
            if verbose:
                print('[%04d - %04d] ep = %05.2f, sh= %09.6f, cl = %08.3f, pn = %05.1f'
                      % (i, j, ep[i], sh[i], cl[j], tmp))
            if (tmp < pn) and ~np.isinf(tmp):
                jn = min(j+1, n)
                cl_out[i] = cl[j]
                break
        if (jn == 1) and (tmp > pn) and ~np.isinf(tmp):
            ep, sh, cl_out = ep[i], sh[i], cl_out[i]
            break
    # Question: why is ep returned in eps and ep?
    return {'eps':ep, 'ep':ep, 'sh':sh, 'cl':cl_out}



import numpy as np
from scipy.signal import chirp, fftconvolve


def pulse_compression(chirp_signal, measured_signal, norm=False):
    """
    Perform pulse compression on a complex signal using a given complex chirp signal with numpy.correlate.
    
    Args:
        signal (numpy.ndarray): The complex signal to be compressed.
        chirp_signal (numpy.ndarray): The complex chirp signal used for compression.
        normalize (bool, optional): Whether to normalize the compressed signal. Default is False.

    Returns:
        numpy.ndarray: The compressed signal (complex).
    """
    # Signals
    c = chirp_signal
    s = measured_signal
    N = len(s)
    
    # Compression
    S = np.fft.fft(s)
    Cconj= np.conj(np.fft.fft(c))
    OUT = S*Cconj
        
    
    # Reformat
    out = np.fft.ifft(OUT)
    out = np.roll(out, int(len(out)/2))
    
    if norm:
        chirp = np.fft.ifft(np.fft.fft(c)*Cconj)
        out = out/np.max(chirp)
    
    return out


#!/usr/bin/env python3

import sys
import unittest
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

sys.path.insert(1, '..')

import subradar as sr


def make_radargram(shape, srfy, walkstd, srfampl, bgampl, rng=None):
    """ Generate a fake radargram with specified dimensions,
    a surface return at, on average, surface location y, with amplitude
    srfampl, and a background amplitude given by bgampl
    shape[0] is the size of the slow time axis and
    shape[1] is the fast time axis """
    assert len(shape) == 2, "Must be a 2D array"
    assert shape[0] > 0 and shape[1] > 0, "Invalid shape"

    if rng is None:
        rng = np.random.default_rng()

    rdg = np.abs(rng.normal(loc=0., scale=bgampl, size=shape))

    # surface walk
    walk = rng.normal(loc=0., scale=walkstd, size=(shape[0],))
    srfloc = srfy + np.cumsum(walk)
    # Round to nearest whole fast time index
    srfloc = np.rint(np.clip(srfloc, 0, shape[1] - 1))

    srfampls = rng.normal(loc=srfampl, scale=srfampl/50., size=(shape[0],))
    for ii in range(shape[0]):
        rdg[ii, int(srfloc[ii])] = srfampls[ii]

    return rdg


class TestClassdef(unittest.TestCase):
    """ Instantiate things in the Classdef file """
    def test_roughness(self):
        x = sr.Classdef.Roughness()
    def test_fresnel(self):
        x = sr.Classdef.Fresnel()
    def test_signal(self):
        x = sr.Classdef.Signal()

class TestIEM(unittest.TestCase):
    """ Test classes/functions in iem.py module """

    def run_common_methods(self, c):
        x = c.f()
        x = c.ft()

    def test_common(self):
        c = sr.iem.Common()
        self.run_common_methods(c)

    def test_large_s(self):
        c = sr.iem.Large_S()

    def test_large_m(self):
        c = sr.iem.Large_M()

    def test_small_m(self):
        c = sr.iem.Small_M()

    def test_small_s(self):
        c = sr.iem.Small_S()
        self.run_common_methods(c)
        x = c.F_sum()
        x = c.Ft()

        for n in range(5):
            x = c.I(n)
            x = c.It(n)

        for db in (False, True):
            for tx in (False, True):
                x = c.nRCS(db=db, transmission=tx)
                y = c.nRCS(db=db, transmission=tx, n=5.1) # valid value for n?



class TestSurface(unittest.TestCase):
    """ Test functions in the surface module """
    def setUp(self):
        self.rng = np.random.default_rng(seed=0xbeef)
        shape = (1000, 800)
        self.rdg = 0.01 + make_radargram(shape, 200, walkstd=2,
                                  srfampl=100, bgampl=0.1, rng=self.rng)

        self.rdg_noise = 0.01 + np.random.normal(0.0, 0.1, size=shape)

    def test_detect_estimate(self):
        """ Test detection using an initial estimate (a good one and a bad one)
        """
        for method in ('grima2012', 'mouginot2010'):
            for y0 in (200, 500):
                with self.subTest(method=method, y0=y0):
                    y0arr = np.full(self.rdg.shape[0:1], y0)
                    x = sr.surface.detector(self.rdg, y0=y0arr, method=method)
                    self.assertEqual(self.rdg.shape[0], len(x))


    def test_detect_method(self):
        for method in ('grima2012', 'mouginot2010'):
            with self.subTest(method=method):
                x = sr.surface.detector(self.rdg, method=method)
                self.assertEqual(self.rdg.shape[0], len(x))

    def test_zeroes(self):
        """ Test with a radargram that is all zeros """
        rdg = np.zeros_like(self.rdg)
        for method in ('grima2012', 'mouginot2010'):
            with self.subTest(method=method):
                x = sr.surface.detector(rdg, method=method)
                self.assertEqual(rdg.shape[0], len(x))

    def test_nosurface(self):
        """ Test with a radargram with no surface in it """
        for method in ('grima2012', 'mouginot2010'):
            with self.subTest(method=method):
                x = sr.surface.detector(self.rdg_noise, method=method)
                self.assertEqual(self.rdg_noise.shape[0], len(x))


    def test_transpose(self):
        rdg = self.rdg.T.copy()
        x = sr.surface.detector(rdg, axis=1)


    def test_gcc(self):
        for method in 'standard wiener roth scot phat ml'.split():
            with self.subTest(weight=method):
                info = sr.surface.gcc(self.rdg, weight=method)

        with self.assertRaises(ValueError):
            info = sr.surface.gcc(self.rdg, weight='unknown!')


class TestRoughness(unittest.TestCase):
    """ test the roughness module """
    def test_isotropic_gaussian(self):
        x = sr.roughness.spectrum(0., 0., 0., 0.)
    def test_unknown(self):
        with self.assertRaises(ValueError):
            x = sr.roughness.spectrum(0., 0., 0., 0., kind='unknown')


class TestUtils(unittest.TestCase):
    """ Test utils that don't otherwise get run """
    def test_geo_loss(self):
        h = np.arange(1, 1000, 0.5)
        y = sr.utils.geo_loss(h)

    def test_funcs1(self):
        x = np.arange(-2*np.pi, 2*np.pi, 0.01)
        y = sr.utils.theta_t2i(x, 1.0, 1.0)
        y = sr.utils.theta_t2i(x, 1.5, 0.5)
    def test_funcs2(self):
        x = np.arange(-np.pi, np.pi, 0.01)
        y = sr.utils.wl2wk(x)
        y = sr.utils.wl2wf(x)
        y = sr.utils.deg2rad(sr.utils.rad2deg(x))
        

def main():
    _ = make_radargram((10, 10), 200, walkstd=2,
                         srfampl=100, bgampl=0.1, rng=None)

    unittest.main()
    #fig, ax = plt. subplots(2, 1, samex=True)
    #ax[0].imshow(rdg.T)
    #ax[1].plot(x)
    #plt.show()

if __name__ == "__main__":
    main()



"""Common utility functions and radar relationships"""

#from numpy import arcsin, cos, pi, sin, sqrt
import numpy as np
import scipy.constants as ct
import scipy.signal
import scipy.fft


#-------------------
# Common conversions
#-------------------


def deg2rad(x):
    """Degree to Radian"""
    return x*np.pi/180

def rad2deg(x):
    """Radian to Degree"""
    return x*180/np.pi

def wf2wl(x):
    """Wavelength to frequency"""
    return ct.c/x

def wl2wf(x):
    """Wavelength to frequency"""
    return ct.c/x

def wl2wk(x):
    """Wavelength to wave number"""
    return 2*np.pi/x

def wf2wk(x):
    """Frequency to wave number"""
    return 2*np.pi/wf2wl(x)

def wk2vec(wk, th):
    """Scalar to Vectorial wave number"""
    return {'x':wk*np.sin(th), 'z':wk*np.cos(th)}


#---------------------
# Fresnel/Snell's Laws
#---------------------


def theta_i2t(x, n1, n2):
    """Incident to transmission angle conversion"""
    return np.arcsin(np.sin(x)*n1/n2)

def theta_t2i(x, n1, n2):
    """Incident to transmission angle conversion"""
    return np.arcsin(np.sin(x)*n2/n1)

def epmu2n(ep, mu):
    """Relative electric perm. and magnetic perm. to optical index"""
    return np.sqrt(ep*mu)

def R_v(ep1, ep2, mu1, mu2, xi):
    """Vertical (parallel) reflection coeff, funct. of incident angle"""
    n1 = epmu2n(ep1, mu1)
    n2 = epmu2n(ep2, mu2)
    xt = theta_i2t(xi, n1, n2)
    z1, z2 = np.sqrt(mu1/ep1), np.sqrt(mu2/ep2)
    return (z2*np.cos(xt) - z1*np.cos(xi)) / (z2*np.cos(xt) + z1*np.cos(xi))

def T_v(ep1, ep2, mu1, mu2, xi):
    """Vertical (parallel) transmission coeff, funct. of incident angle"""
    n1 = epmu2n(ep1, mu1)
    n2 = epmu2n(ep2, mu2)
    xt = theta_i2t(xi, n1, n2)
    z1, z2 = np.sqrt(mu1/ep1), np.sqrt(mu2/ep2)
    return 2*z2*np.cos(xi) / (z2*np.cos(xt) + z1*np.cos(xi))

def R_h(ep1, ep2, mu1, mu2, xi):
    """Horizontal (perpandicular) reflection coeff, funct. of incident angle"""
    n1 = epmu2n(ep1, mu1)
    n2 = epmu2n(ep2, mu2)
    xt = theta_i2t(xi, n1, n2)
    z1, z2 = np.sqrt(mu1/ep1), np.sqrt(mu2/ep2)
    return (z2*np.cos(xi) - z1*np.cos(xt)) / (z2*np.cos(xi) + z1*np.cos(xt))

def T_h(ep1, ep2, mu1, mu2, xi):
    """Horizontal (perpandicular) transmission coeff, funct. of inc. angle"""
    n1 = epmu2n(ep1, mu1)
    n2 = epmu2n(ep2, mu2)
    xt = theta_i2t(xi, n1, n2)
    z1, z2 = np.sqrt(mu1/ep1), np.sqrt(mu2/ep2)
    return 2*z2*np.cos(xi) / (z2*np.cos(xi) + z1*np.cos(xt))

def R(ep1, ep2, mu1, mu2, xi):
    """Unpolarized Reflection coefficient"""
    return (R_v(ep1, ep2, mu1, mu2, xi) + R_h(ep1, ep2, mu1, mu2, xi))/2.


#---------------------
# Footprints
#---------------------


def footprint_rad_beam(h, bmw):
    """radius of the beam-limited footprint"""
    return np.abs(h)*bmw/2.

def footprint_rad_pulse(h, wb):
    """radius of the pulse-limited footprint"""
    return np.sqrt(np.abs(h)*ct.c/wb)

def footprint_rad_fresnel(h, wl):
    """radius of the Fresnel footprint"""
    return np.sqrt(np.abs(h)*wl/2)


#------
# Misc.
#------

def geo_loss(h):
    """Energy losses from geometric spreading"""
    return 1/(4*np.pi*h**2)


#-------
# Others
#-------

#def coherence(x, y, n=None):
#    if n is None:
#        n = x.size
#    _, Cxy = scipy.signal.coherence(x, y, nfft=n)
#    return Cxy

def gcc(x, y, weight='standard', **kwargs):
    """Generalized Cross-correlation between two vectors

    Arguments
    ---------
    x: vector array
        signal 1
    y: vector array
        signal 2
    weight: string
        filter to be applied on both signals

    Output
    ------
    tau: integer
        time delay
    val: float
        Maximum value of the cross-corelation
    cc: array
        Correlation Coefficient
    ch: array
        Magnitude squared coherence of x and y.
    """
    n_ini = int(x.size)
    _, ch = scipy.signal.coherence(x, y, nfft=n_ini)
    x, y = np.abs(x), np.abs(y)
    #padx = np.zeros(x.size)
    #pady = np.zeros(y.size)
    x = np.hstack([x,x])
    y = np.hstack([y,y])
    n = x.size
    X = scipy.fft.fft(x)
    Y = scipy.fft.fft(y)
    Pxx = X * np.conj(X)
    Pyy = Y * np.conj(Y)
    Pxy = X * np.conj(Y)

    Cxy = ch
    if weight == 'standard':
        W = 1.
    elif weight == 'wiener':
        W = 1./Cxy
    elif weight == 'roth':
        W = 1./Pxx
    elif weight == 'scot':
        W = 1./np.sqrt(Pxx * Pyy)
    elif weight == 'phat':
        W = 1./np.abs(Pxy)
    elif weight == 'ml':
        K = Cxy / (1.-Cxy)
        W =  K / np.abs(Pxy)
    else:
        raise ValueError("Unknown weight '%s'" % weight)

    # Correlation Coefficient
    cc = np.abs(scipy.fft.ifft(Pxy * W ))#[1:]
    n_ini = int(n/2)
    cc = cc[n_ini:n_ini+n]
    cc = np.roll(cc, n_ini)

    # TDE
    tau = np.abs(cc).argmax()
    tau = np.mod(tau, n_ini).astype(int)
    tau = tau if tau <= n_ini/2 else tau-n_ini
    val = np.abs(cc).max()

    # TODO: this would be somewhat quicker as a namedtuple
    return {'tau':tau, 'val':val, 'cc':cc, 'ch':ch}



import os
import numpy as np
import scipy
import glob
#import rsr
import matplotlib.pyplot as plt
import h5py
import hdf5storage
from .. import filtering, surface


class Results:
    """ Class to read results as generated by the Trento simulator [Gerekos et al., 2018]
    """
    def __init__(self, 
                 simulation_name: str, # Folder name of the simulation
                 source_folder=('..', 'data', 'RadarSimulationResults'),
                 gain=1):
        
        self.simulation_name = simulation_name
        self.output_folder = os.path.join(*source_folder, simulation_name, 'OUTPUTS')
        self.input_folder = os.path.join(*source_folder, simulation_name, 'INPUTS')
        self.input_filenames = glob.glob(os.path.join(self.input_folder, '*.mat'))
        self.gain = gain
        self.trajectory = self.read_trajectory()
        self.simulation = self.read_simulation()
        self.geoelectric = self.read_geoelectric()
        #self.inputs = dict()
        #_ = [self.inputs.update((self.loadmat(filename)) for filename in self.input_filenames]
        self.inputs = {**self.trajectory, **self.simulation, **self.geoelectric}
    
    
    def frame_filenames(self):
        """ Provide the list of frame filenames 
        """
        out = glob.glob(os.path.join(self.output_folder, '*.mat'))
        out = np.sort(out)
        return out
    
    
    def loadmat(self, filename):
        """Load matfile
        """
        mat = hdf5storage.loadmat(filename)
        out = dict()
        for key in mat.keys():
            data = mat[key][0]
            if np.shape(data)[0] == 1:
                data = data[0]
            out[key] = data
        return out
    
            
    def read_simulation(self):
        filename = [filename for filename in self.input_filenames if 'Simulation_' in filename][0]
        out = self.loadmat(filename)
        return out
    
    
    def read_geoelectric(self):
        filename = [filename for filename in self.input_filenames if 'Geoelectrical_' in filename][0]
        out = self.loadmat(filename)
        
        # Add the DEM (for any reason, only h5py sees it, not hdf5storage)
        f = h5py.File(filename)
        for key in f.keys() :
            if 'group' in f[key].__str__():
                for dset in f[key].keys():  
                    ds_data = f[key][dset] # returns HDF5 dataset object
                    arr = f[key][dset][:] # adding [:] returns a numpy array
                    out[dset] = arr
        
        return out
    
    
    def read_trajectory(self):
        filename = [filename for filename in self.input_filenames if 'Trajectory_' in filename][0]
        out = self.loadmat(filename)
        return out
    
    
    def dem(self):
        """Provide the input DEM
        """
        out = self.inputs['b']
        out = np.rot90(out, k=1)
        return out
    
    
    def xy(self, meters=True):
        """Provide [x, y] coordinates of the frames
        """
        #x0_meter = self.input_data('Traj')['sc_position_x']
        #y0_meter = self.input_data('Traj')['sc_position_y']
        dx = self.inputs['DELTA_X']
        #length = int(self.input_data('Traj')['TrackLengthTot'])
        
        xs = self.inputs['sc_position_x']
        ys = self.inputs['sc_position_y']
        
        #xs = np.full(length, x0_meter)#np.arange(length)*dx + x0_meter
        #ys = np.arange(length)*dx + y0_meter
        
        if not meters:
            # Values in pixels
            xs = xs/dx
            ys = ys/dx
        
        return xs, ys
    
    
    def plot_trajectory(self):
        # Parameters
        dem = self.dem()
        dx = self.inputs['DELTA_X']
        x, y = self.xy(meters=True)
        radius = self.inputs['PulseLtdR']
        x_length = self.inputs['l']*dx
        y_length = self.inputs['m']*dx

        # FIGURE
        # ------
        fig, ax = plt.subplots()
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')
        ax.set_title(self.simulation_name)
        
        # DEM plot
        #im = ax.imshow(dem, extent=[0,x_length*dx,0,y_length*dx])
        im = ax.imshow(dem, extent=[y_length,0,x_length,0])

        # colorbar
        cax = fig.add_axes([ax.get_position().x1+0.01,ax.get_position().y0,0.02,ax.get_position().height])
        plt.colorbar(im, cax=cax, label='Elevation [m]') # Similar to fig.colorbar(im, cax = cax)
        
        # Trajectory plot
        ax.plot(x, y, 'k',)

        # Footprint plot
        for i in [0, -1]:
            circle = plt.Circle((x[i],y[i]), radius, fill = False , color='k', ls='--')
            ax.set_aspect( 1 )
            ax.add_artist( circle )
            
        return ax
        

    def plot_radargram(self, aspect:int=6, microsec:bool=True, **kwargs):
        """Browse product of the results
        kwargs = Anything to pass to read_frame (including compression options
        """
        fast_time = self.inputs['t']
         
        fig, ax = plt.subplots()
        
        rdg = self.radargram(rotate=True, **kwargs)
        
        if microsec:
            extent = [0, np.shape(rdg)[1], fast_time[-1]*1e6, fast_time[0]*1e6]
        else:
            extent = [0, np.shape(rdg)[1], 0, np.shape(rdg)[0]]
            
        im = ax.imshow(rdg, extent=extent, aspect=aspect)
        
        # colorbar
        cax = fig.add_axes([ax.get_position().x1+0.01,ax.get_position().y0,0.02,ax.get_position().height])
        plt.colorbar(im, cax=cax, label='Power [dB]') # Similar to fig.colorbar(im, cax = cax)
        
        ax.set_title(self.simulation_name)
        ax.set_ylabel(r'Time [$\mu$s]')
        ax.set_xlabel(r'Range bin')
        
        return ax
    
    
    def plot_frame(self, i):
        """plot a frame
        i = frame number
        """
        # Signal
        raw_signal = self.read_frame(i)
        nowin_signal = self.read_frame(i, compression='No windowing')
        win_signal = self.read_frame(i, compression='Hann windowing')
        time = self.inputs['t']*1e6

        # Plot
        fig, axs = plt.subplots(1, 2, figsize=(15, 4))

        ax = axs[0]
        # Plot the measured signal in time domain
        ax.plot(time, np.absolute(raw_signal), lw=.5, label='Raw')
        ax.set_title('Measured Signal in Time Domain')
        ax.set_xlabel(r'Time [$\mu$s]')
        ax.set_ylabel('Amplitude')
        ax.grid(True)

        ax = axs[1]
        # Plot the compressed signal in time domain
        ax.plot(time, np.absolute(win_signal), label='No windowing')
        # Plot the compressed signal in time domain
        ax.plot(time, np.absolute(win_signal), label='Hann windowing')
        ax.set_title('Compressed Signal in Time Domain')
        ax.set_xlabel(r'Time [$\mu$s]')
        ax.set_ylabel('Amplitude')
        ax.grid(True)
        ax.legend()
        
        return axs
    
    
    def read_frame(self, i:int,
                   compression:str='Hann windowing', 
                   absolute:bool=False, 
                   norm=False,
                   pdb:bool=False,
                   **kwargs):
        """Read a frame (range line)
        i = frame number
        compression = "No windowing" or "Hann windowing"
        absolute = Provide the absolute values instead of complex number
        norm = Whether to normalize to the max of the chirp
        pdb = Provide power in dB
        """
        i = i-1
        out = scipy.io.loadmat(self.frame_filenames()[i])['Final_signal'][0]
        out = out*self.gain
        
        if compression == 'Hann windowing':
            chirp = self.inputs['Signal']/self.inputs['Norm']
            out = filtering.pulse_compression(chirp, out, norm=norm)
            
        elif compression == 'No windowing':
            chirp = self.inputs['Signal_raw']/self.inputs['Norm']
            out = filtering.pulse_compression(chirp, out, norm=norm)
        else:
            pass
        
        if absolute:
            out = np.absolute(out)
            
        if pdb:
            out = 20*np.log10(np.absolute(out))
            
        return out
        
        
    
    def radargram(self, rotate=False, **kwargs):
        """ Stack the range lines to produce a 2D radargram
        rotate = rotate the array by 90 degrees
        kwargs = Anything to pass to read_frame
        """
        Nx = len(self.frame_filenames())
        Ny = len(self.inputs['signal_window'])
        
        for i in np.arange(Nx):
            frame = self.read_frame(i, **kwargs)
            # Make results array of correct type
            if i == 0:
                if frame.dtype == 'complex128':
                    out = np.empty((Nx, Ny), dtype='complex128')
                else:
                    out = np.empty((Nx, Ny))
            out[i,:] = frame
        
        if rotate:
            out = np.rot90(out)
            
        return out
        
        
    def surface(self, **kwargs):
        """Surface echo extraction
        kwargs = Anything to pass to read_frame
        """
        if 'compression' in kwargs:
            compression = kwargs['compression']
        else:
            compression = None
            
        rdg_ref = self.radargram(compression=compression, absolute=True)
        rdg = self.radargram(**kwargs)
        
        ys = [np.argmax(rdg_ref[i,:]) for i in np.arange(np.shape(rdg)[0])]
        val = [rdg[i,ys[i]] for i in np.arange(len(ys))]
        
        # Spatial Coordinates
        coord = self.xy()
        
        return {'y':ys, 'val':val, 'coord':coord}
        
        
#    def surface_rsr(self, fit_model='hk', lim=None, **kwargs):
#        """RSR
#        """
#        if 'compression' in kwargs:
#            compression = kwargs['compression']
#        else:
#            compression = None
            
#        srf = self.surface(compression=compression, absolute=True)
#        amp = np.array(srf['val'])
        
#        if lim:
#            y = srf['coord'][1]
#            idx = [i for i, val in enumerate(y) if (val > lim[0]) and (val < lim[1])]
#            amp2 = [amp[i] for i in idx]
#            amp = np.array(amp2)
            
#        f = rsr.run.processor(amp, fit_model=fit_model)
#        return f


"""Various tools for surface detection"""

import numpy as np
import pandas as pd
import scipy.signal
from . import utils


def detector(rdg, y0=None, winsize=100, method='grima2012', axis=0, out_dict=False, **kwargs):
    """Surface detection with the chosen method

    Input
    -----
    rdg: 2d-array
        radargram.

    y0: 1d-array
        Initial estimation for the location of the surface.
        Optional.

    winsize: float
        Size of the window around y0 to look for the surface.
        Activated only if y0 is a nonempty array.

    method: string
        method to use for surface detection.

    axis: 0 or 1
        Indicate which axis is long-time (slow-time) axis. default=0

    out_dict: True or False
        Return a dictionary

    Output
    ------
    y: float
        index of the location of the detected echo.

    """
    if axis == 1:
        rdg = np.transpose(rdg)

    xsize, ysize = rdg.shape
    y = np.empty(xsize)
    c = np.empty(xsize)

    detectors = {
        'maximum': maximum,
        'grima2012': grima2012,
        'mouginot2010': mouginot2010,
    }
    fdetector = detectors[method]

    # Detection
    for xi in range(xsize):
        #index vector
        if y0 is not None and len(y0) > 0:
            idx = np.arange(winsize)+y0[xi]-winsize/2.
        else:
            idx = np.arange(ysize)

        y[xi], cs = fdetector(rdg[xi, :], idx=idx, **kwargs)
        c[xi] = np.nanmax(cs)

    if out_dict:
        return {'y':y, 'c':c}
    else:
        return y


def maximum(signal, idx=(), **kwargs):
    """Surface detection from the maximum within a window
    
    Parameters
    ----------
    signal: array
        signal vector

    idx: array
        the indices of the array where to search for the echo

    Return
    ------
    y: float
        index of the location of the detected echo

    c: array
        criteria computed with idx
    """

    # array of index where to search for the surface
    idx = np.array(idx)
    if idx.size == 0 :
        idx = np.arange(len(signal)).astype(int)
    else:
        idx = np.array(idx).astype(int) # make idx an integer array

    # Estimator calculation
    c = signal[idx]

    # surface index
    try:
        y = idx[np.nanargmax(c)]
    except ValueError:
        y = np.nan
    return y, c


def mouginot2010(signal, idx=(), period=3, window=30, **kwargs):
    """Surface detection using [Mouginot et al. 2010]

    Parameters
    ----------
    signal: array
        signal vector

    idx: array
        the indices of the array where to search for the echo

    period: float
        window shift to compute the noise (=1 in the original paper)

    window: float
        size of the window where to compute the noise

    Output
    ------
    y: float
        index of the location of the detected echo

    c: array
        criteria computed with idx
    """

    # array of index where to search for the surface
    idx = np.array(idx)
    if idx.size == 0 :
        idx = np.arange(len(signal)).astype(int)
    else:
        idx = np.array(idx).astype(int) # make idx an integer array

    # Estimator calculation
    noise = pd.Series(signal[idx]).shift(periods=period).rolling(window).mean().values
    #noise = [np.nanmean(signal[i-30:i-3]) for i in idx]
    c = signal[idx]/noise

    # surface index
    try:
        y = idx[np.nanargmax(c)]
    except ValueError:
        y = np.nan
    return y, c


def grima2012(signal, idx=(), **kwargs):
    """Surface detection from [Grima et al. 2012]

    Parameters
    ----------
    signal: array
        signal vector

    idx: array
        the indices of the array where to search for the echo

    Return
    ------
    y: float
        index of the location of the detected echo

    c: array
        criteria computed with idx
    """

    # array of index where to search for the surface
    idx = np.array(idx)
    if idx.size == 0 :
        idx = np.arange(len(signal)).astype(int)
    else:
        idx = np.array(idx).astype(int) # make idx an integer array

    # Estimator calculation
    s1 = signal[idx]
    derivative = np.roll(np.gradient(s1), 2)
    c = s1*derivative

    # surface index
    try:
        y = idx[np.nanargmax(c)]
    except ValueError:
        y = np.nan
    return y, c


def gcc(rdg, tau_threshold=2, **kwargs):
    """Surface detection from relative time delay obtained through generalized
    cross-correlation of each contiguous range lines

    Parameters
    ----------
    rdg: 2d-array
        radargram
        axis 0 is slow time, axis 1 is fast time

    Return
    ------
    """
    #---------------
    # Initialization
    xsize, ysize = rdg.shape
    yn = list(range(ysize))
    tau = np.empty(ysize, dtype=int)
    val = np.empty(ysize)
    cc = np.empty_like(rdg) #np.abs(rdg)*0
    #ch = np.empty_like(rdg) #np.abs(rdg)*0
    #offset = np.empty(ysize, dtype=int)

    #-------------------------
    # GCC applied on radargram

    # All records except last
    for i in range(ysize-1):
        _ = utils.gcc(rdg[:, i], rdg[:, i+1], **kwargs)
        tau[i] = _['tau']
        val[i] = _['val']
        cc[:,i] = _['cc']
        #ch[:,i] = _['ch']

    # Question: can we reuse the last value calculated?
    # Last record
    _ = utils.gcc(rdg[:, i], rdg[:, i-1], **kwargs)
    tau[-1] = _['tau']
    val[-1] = _['val']
    cc[:,-1] = _['cc']
    #ch[:,-1] = _['ch']

    # Quality flag when tau gradient higher than dtau_threshold
    #dtau = np.roll( np.gradient( np.abs(tau)) ,-1)
    ok = np.abs(tau) <= tau_threshold


    #----------------------------------------
    # Vertical offset that correponds to tau

    offset = np.cumsum(tau) # tau is an int, so offset should be, too


    #-------------------
    # Corrected offsets

    #Radargram rolled with offset
    rdg2 = np.empty_like(rdg)
    for i in range(ysize):
        rdg2[:,i] = np.roll(rdg[:,i], offset[i])

    # TODO: does this really need to be a nested function?
    # Radargram is divided by chunks that are bounded where ok=0
    def _data_chunks(data, stepsize=1):
        data_id = np.arange(data.size)*data
        pieces = np.split(data_id, np.where(np.diff(data_id) != stepsize)[0]+1)
        chunks = [i for i in pieces if (i.size > 1)]
        return [np.array(chunk, dtype=int) for chunk in chunks]

    chunks = _data_chunks(ok)

    # Cumulative sum of each chunk to assess the average coordinate
    # of the surface echo in each chunk
    chunk_cumsums = [np.abs(rdg2[:, chunk].sum(axis=1)) for chunk in chunks]
    chunk_cumsum_argmaxs = [np.argmax(chunk_cumsum) for chunk_cumsum in chunk_cumsums]

    # Chunks are aligned for their average surface echo coordinate to match
    offset2 = np.empty_like(offset)
    for i, chunk in enumerate(chunks):
        offset2[chunk] = offset[chunk] - chunk_cumsum_argmaxs[i] + chunk_cumsum_argmaxs[0]

    del rdg2
    # NB: if you're doing a del, you should probably make this a function

    #-------------------------------
    # Coordinate of the surface echo

    rdg3 = np.empty_like(rdg)
    for i in range(ysize):
        rdg3[:,i] = np.roll(rdg[:,i], offset2[i])
    y0 = np.argmax( np.abs(rdg3.sum(axis=1)) )
    y = y0 + offset2
    del rdg3

    # TODO: remove yn?  do you really need it?
    return {'tau':tau.astype(int), 'val':val, 'cc':cc, 'ok':ok, 'yn':yn,
    'offset':offset, 'offset2':offset2, 'y':y}
